#! /usr/bin/python

from twisted.python.components import Interface


class ITransport(Interface):
    """An object used to get a message blob to the receiver. Each ITransport
    handles a single message to a single recipient. It will exist until the
    message has been received safely (if an acknowledgement is expected,
    this object will stick around until after the ack has been received).

    The object is created with a specific recipient in mind.
    """

    def __init__(self, args):
        """
        @type args: C{string}
        @param args: this will have the text after the prefix:
        """

    def send(self, encapsulated):
        """Queue a message for delivery. Returns a Deferred which fires when
        the message is out of our hands.

        @type encapsulated: C{string}
        @param encapsulated: the encrypted message blob
        """

class IInlet(Interface):
    """An object used to accept incoming messages from remote senders. The
    recipient specifies one of these and tells the rest of the world to use
    it.

    Upon receipt of a message, self.service.receiveMessage(encrypted) should
    be called.

    This object should sensibly handle being serialized.
    """

    def start(self):
        """Called when the inlet is first activated (when it is added to the
        application). This is an appropriate time to open listening sockets,
        start timers to poll mailboxes, etc.
        """

    def shutdown(self):
        """Called when the inlet is being removed. Sockets should be closed,
        timers should be canceled.
        """

    def encode(self):
        """Should return a string which describes this transport to the rest
        of the world. This string will be put into the Identity Record.
        """

class IMessage(Interface):
    """This is an outbound message."""

    def provideFeature(self, name):
        """A Processor uses this to tell me that it has provided a feature,
        which may allow other processors to be executed."""

class IReceivedMessage(Interface):
    """This is an inbound message."""
    pass

class IAddressBook(Interface):
    pass
class IAddressBookEntry(Interface):
    pass
class IOwnerEntry(Interface):
    pass
class IIDRecord(Interface):
    pass
class ITicketAgent(Interface):
    """When an outgoing message needs a ticket (msg.requirements is
    non-empty), these agents are asked to satisfy these requirements. This
    usually involves:

     locate and talk to a suitable Ticket Vendor
     presenting their challenge to the user
     conveying the user's response back to the vendor
     accept the vendor's ticket
     add the ticket to the message
     remove the msg.requirements entry
    """

    def handleRequirements(self, msg):
        """Try to satisfy msg.requirements. Should remove all requirements
        thus satisfied before returning.
        """

class IOwnerPermission(Interface):
    """I describe someone else's permission to send a message to me. I am
    held by the receiver.
    """

class IPermissionSlip(Interface):
    """I describe my permission to send a message to someone else. I am held
    by the sender."""

class IProcessor(Interface):
    """I examine messages. Each agent has a list of Processors. Many are
    built-in, but others can come from plugins. Each Processor lives in a
    single agent.
    """

    def setAgent(self, agent):
        """Bind this Processor to the agent it lives in."""

    def getDependencies(self):
        """Return a list of strings, each of which represents a dependency
        of this processor. The Processor will not be run until the
        dependency is fulfilled. Note that it may be fulfilled negatively
        (indicating the required data will never be available)."""

    def getProvides(self):
        """Return a list of strings which indicate all the features which
        this processor will eventually provide (by calling message.provide
        with the feature name).
        """

    def configureSender(self, sender):
        """Hazy: Create a dialog that can be used to set sender-global state
        for this Processor."""

    def configureRecipient(self, recipient):
        """Hazy: add to the address-book page for a single recipient,
        allowing configuration of per-recipient Processor state."""

class IOutboundProcessor(IProcessor):
    """I sign off on outbound messages, and have the ability to modify,
    delay, or reject each one. In general, I will perform a lookup and put
    the results into header nodes inside the Message.

    I will not be asked until all my dependencies are met.

    I will only be asked about the message once. Once all Processors agree,
    the message will be delivered.

    well-known dependencies:
     recipientRecord: fulfilled when message.recipientRecord is either
                      non-None or will never become non-None (?). For messages
                      which are sent using the address book, messages sent to
                      an explicit IDRecord, replies, and most control messages,
                      this will be pre-fulfilled. For messages sent to an
                      Address, this will be fulfilled by the AddressLookup
                      processor.
                      
     recipientEntry: fulfilled when message.recipientEntry is either
                     a valid AddressBookEntry or will never be non-None. For
                     messages sent using the address book, it will be
                     pre-fulfilled. For entries sent to an Address, it will
                     be fulfilled by AddressLookup when it creates the
                     AddressBookEntry.

                     recipientEntry is dependent upon recipientRecord.

     
    """    

    def checkOutbound(self, message):
        """Do stuff to an outbound Message.

        I will not be called until all the features in my .getDependencies
        list are provided (probably by other processors).

        I return True if I am willing to let this message be delivered now.

        I return a Deferred if I have initiated a lookup or other
        asyncronous processing. I will fire the Deferred with 'True' when
        that processing is finished and I am willing to let the message
        pass. I will provide all dependencies that I can before firing the
        deferred.

        If I will refuse to let the message through no matter what the other
        Processors eventually say, I will raise a Unsendable exception. I
        may also so this by firing the deferred's errback.
        """

class IInboundProcessor(IProcessor):
    """For inbound messages, I will look at the Message and extract data
    from it. This will continue until all Processors have indicated that
    they have no more data to extract."""

    def checkInbound(self, receivedMessage):
        """Do stuff to an inbound ReceivedMessage.

        I return True when I am done extracting information from the
        message. Once I return True, I will not be asked about the message
        again.

        I return False if I am not yet finished extracting information
        (perhaps I am waiting for some other Processor to give me a place to
        put it).
        """

class IRequest(Interface):
    pass

class Unsendable(Exception):
    pass
class NotPermitted(Exception):
    """This incoming message was rejected"""
    pass
